#!/usr/bin/env ruby
# encoding: utf-8

module Coopy
  class CombinedTable 
    
    def initialize(t)
      @t = t
      @dx = 0
      @dy = 0
      @core = t
      @head = nil
      return if t.get_width < 1 || t.get_height < 1
      v = t.get_cell_view
      return if v.to_s(t.get_cell(0,0)) != "@@"
      @dx = 1
      @dy = 0
      begin
        _g1 = 0
        _g = t.get_height
        while(_g1 < _g) 
          y = _g1
          _g1+=1
          txt = v.to_s(t.get_cell(0,y))
          break if txt == nil || txt == "" || txt == "null"
          @dy+=1
        end
      end
      @head = ::Coopy::CombinedTableHead.new(self,@dx,@dy)
      @body = ::Coopy::CombinedTableBody.new(self,@dx,@dy)
      @core = @body
      @meta = ::Coopy::SimpleMeta.new(@head)
    end
    
    protected
    
    attr_accessor :t
    attr_accessor :body
    attr_accessor :head
    attr_accessor :dx
    attr_accessor :dy
    attr_accessor :core
    attr_accessor :meta
    
    public
    
    def all 
      @t
    end
    
    def get_table 
      self
    end
    
    def height() get_height end
    def height=(__v) @height = __v end
    def width() get_width end
    def width=(__v) @width = __v end
    
    def get_width 
      @core.get_width
    end
    
    def get_height 
      @core.get_height
    end
    
    def get_cell(x,y)
      @core.get_cell(x,y)
    end
    
    def set_cell(x,y,c)
      @core.set_cell(x,y,c)
    end
    
    def to_s 
      ::Coopy::SimpleTable.table_to_string(self)
    end
    
    def get_cell_view 
      @t.get_cell_view
    end
    
    def is_resizable 
      @core.is_resizable
    end
    
    def resize(w,h)
      @core.resize(h,w)
    end
    
    def clear 
      @core.clear
    end
    
    def insert_or_delete_rows(fate,hfate)
      @core.insert_or_delete_rows(fate,hfate)
    end
    
    def insert_or_delete_columns(fate,wfate)
      @core.insert_or_delete_columns(fate,wfate)
    end
    
    def trim_blank 
      @core.trim_blank
    end
    
    def get_data 
      nil
    end
    
    def clone 
      @core.clone
    end
    
    def create 
      @t.create
    end
    
    def get_meta 
      @meta
    end
    
  haxe_me ["coopy", "CombinedTable"]
  end

end
