#!/usr/bin/env ruby
# encoding: utf-8

module Coopy
  class CombinedTableBody 
    
    def initialize(parent,dx,dy)
      @parent = parent
      @dx = dx
      @dy = dy
      @all = parent.all
    end
    
    protected
    
    attr_accessor :parent
    attr_accessor :dx
    attr_accessor :dy
    attr_accessor :all
    attr_accessor :meta
    
    public
    
    def get_table 
      self
    end
    
    def height() get_height end
    def height=(__v) @height = __v end
    def width() get_width end
    def width=(__v) @width = __v end
    
    def get_width 
      @all.get_width - 1
    end
    
    def get_height 
      @all.get_height - @dy + 1
    end
    
    def get_cell(x,y)
      if y == 0 
        @meta = @parent.get_meta.as_table if @meta == nil
        return @meta.get_cell(x + @dx,0)
      end
      @all.get_cell(x + @dx,y + @dy - 1)
    end
    
    def set_cell(x,y,c)
      if y == 0 
        @all.set_cell(x + @dx,0,c)
        return
      end
      @all.set_cell(x + @dx,y + @dy - 1,c)
    end
    
    def to_s 
      ::Coopy::SimpleTable.table_to_string(self)
    end
    
    def get_cell_view 
      @all.get_cell_view
    end
    
    def is_resizable 
      @all.is_resizable
    end
    
    def resize(w,h)
      @all.resize(w + 1,h + @dy)
    end
    
    def clear 
      @all.clear
      @dx = 0
      @dy = 0
    end
    
    def insert_or_delete_rows(fate,hfate)
      fate2 = Array.new
      begin
        _g1 = 0
        _g = @dy
        while(_g1 < _g) 
          y = _g1
          _g1+=1
          fate2.push(y)
        end
      end
      hdr = true
      begin
        _g2 = 0
        while(_g2 < fate.length) 
          f = fate[_g2]
          _g2+=1
          if hdr 
            hdr = false
            next
          end
          fate2.push(((f >= 0) ? f + @dy - 1 : f))
        end
      end
      @all.insert_or_delete_rows(fate2,hfate + @dy - 1)
    end
    
    def insert_or_delete_columns(fate,wfate)
      fate2 = Array.new
      begin
        _g1 = 0
        _g = @dx + 1
        while(_g1 < _g) 
          x = _g1
          _g1+=1
          fate2.push(x)
        end
      end
      begin
        _g2 = 0
        while(_g2 < fate.length) 
          f = fate[_g2]
          _g2+=1
          fate2.push(((f >= 0) ? f + @dx + 1 : f))
        end
      end
      @all.insert_or_delete_columns(fate2,wfate + @dx)
    end
    
    def trim_blank 
      @all.trim_blank
    end
    
    def get_data 
      nil
    end
    
    def clone 
      ::Coopy::CombinedTable.new(@all.clone)
    end
    
    def create 
      ::Coopy::CombinedTable.new(@all.create)
    end
    
    def get_meta 
      @parent.get_meta
    end
    
  haxe_me ["coopy", "CombinedTableBody"]
  end

end
