#!/usr/bin/env ruby
# encoding: utf-8

module Coopy
  class SqlTable 
    
    def initialize(db,name,helper = nil)
      @db = db
      @name = name
      @helper = helper
      @helper = db.get_helper if helper == nil
      @cache = {}
      @h = -1
      @id2rid = nil
      self.get_columns
    end
    
    protected
    
    attr_accessor :db
    attr_accessor :columns
    attr_accessor :name
    attr_accessor :quoted_table_name
    attr_accessor :cache
    attr_accessor :column_names
    attr_accessor :h
    attr_accessor :helper
    attr_accessor :id2rid
    
    def get_columns 
      return if @columns != nil
      return if @db == nil
      @columns = @db.get_columns(@name)
      @column_names = Array.new
      begin
        _g = 0
        _g1 = @columns
        while(_g < _g1.length) 
          col = _g1[_g]
          _g+=1
          @column_names.push(col.get_name)
        end
      end
    end
    
    public
    
    def get_primary_key 
      self.get_columns
      result = Array.new
      begin
        _g = 0
        _g1 = @columns
        while(_g < _g1.length) 
          col = _g1[_g]
          _g+=1
          next if !col.is_primary_key
          result.push(col.get_name)
        end
      end
      result
    end
    
    def get_all_but_primary_key 
      self.get_columns
      result = Array.new
      begin
        _g = 0
        _g1 = @columns
        while(_g < _g1.length) 
          col = _g1[_g]
          _g+=1
          next if col.is_primary_key
          result.push(col.get_name)
        end
      end
      result
    end
    
    def get_column_names 
      self.get_columns
      @column_names
    end
    
    def get_quoted_table_name 
      return @quoted_table_name if @quoted_table_name != nil
      @quoted_table_name = @db.get_quoted_table_name(@name)
      @quoted_table_name
    end
    
    def get_quoted_column_name(name)
      @db.get_quoted_column_name(name)
    end
    
    def get_cell(x,y)
      if @h >= 0 
        y = y - 1
        y = @id2rid[y] if y >= 0
      elsif y == 0 
        y = -1
      end
      if y < 0 
        self.get_columns
        return @columns[x].name
      end
      row = @cache[y]
      if row == nil 
        row = {}
        self.get_columns
        @db.begin_row(@name,y,@column_names)
        while(@db.read) 
          _g1 = 0
          _g = self.get_width
          while(_g1 < _g) 
            i = _g1
            _g1+=1
            begin
              v = @db.get(i)
              begin
                value = v
                row[i] = value
              end
              v
            end
          end
        end
        @db._end
        begin
          @cache[y] = row
          row
        end
      end
      begin
        this1 = @cache[y]
        this1.get(x)
      end
    end
    
    def set_cell_cache(x,y,c)
      row = @cache[y]
      if row == nil 
        row = {}
        self.get_columns
        begin
          @cache[y] = row
          row
        end
      end
      begin
        v = c
        begin
          value = v
          row[x] = value
        end
        v
      end
    end
    
    def set_cell(x,y,c)
      puts "SqlTable cannot set cells yet"
    end
    
    def get_cell_view 
      ::Coopy::SimpleView.new
    end
    
    def is_resizable 
      false
    end
    
    def resize(w,h)
      false
    end
    
    def clear 
    end
    
    def insert_or_delete_rows(fate,hfate)
      false
    end
    
    def insert_or_delete_columns(fate,wfate)
      false
    end
    
    def trim_blank 
      false
    end
    
    def height() get_height end
    def height=(__v) @height = __v end
    def width() get_width end
    def width=(__v) @width = __v end
    
    def get_width 
      self.get_columns
      @columns.length
    end
    
    def get_height 
      return @h if @h >= 0
      -1
    end
    
    def get_data 
      nil
    end
    
    def clone 
      nil
    end
    
    def create 
      nil
    end
    
    def get_meta 
      self
    end
    
    def alter_columns(columns)
      result = @helper.alter_columns(@db,@name,columns)
      @columns = nil
      result
    end
    
    def change_row(rc)
      if @helper == nil 
        puts "No sql helper"
        return false
      end
      if rc.action == "+++" 
        return @helper.insert(@db,@name,rc.val)
      elsif rc.action == "---" 
        return @helper._delete(@db,@name,rc.cond)
      elsif rc.action == "->" 
        return @helper.update(@db,@name,rc.cond,rc.val)
      end
      false
    end
    
    def as_table 
      pct = 3
      self.get_columns
      w = @column_names.length
      mt = ::Coopy::SimpleTable.new(w + 1,pct)
      mt.set_cell(0,0,"@")
      mt.set_cell(0,1,"type")
      mt.set_cell(0,2,"key")
      begin
        _g = 0
        while(_g < w) 
          x = _g
          _g+=1
          i = x + 1
          mt.set_cell(i,0,@column_names[x])
          mt.set_cell(i,1,@columns[x].type_value)
          mt.set_cell(i,2,((@columns[x].primary) ? "primary" : ""))
        end
      end
      mt
    end
    
    def use_for_column_changes 
      true
    end
    
    def use_for_row_changes 
      true
    end
    
    def clone_meta(table = nil)
      nil
    end
    
    def apply_flags(flags)
      false
    end
    
    def get_database 
      @db
    end
    
    def get_row_stream 
      self.get_columns
      @db._begin("SELECT * FROM " + _hx_str(self.get_quoted_table_name) + " ORDER BY ?",[@db.rowid],@column_names)
      self
    end
    
    def is_nested 
      false
    end
    
    def is_sql 
      true
    end
    
    def fetch_row 
      if @db.read 
        row = {}
        begin
          _g1 = 0
          _g = @column_names.length
          while(_g1 < _g) 
            i = _g1
            _g1+=1
            begin
              v = @db.get(i)
              begin
                value = v
                row[@column_names[i]] = value
              end
              v
            end
          end
        end
        return row
      end
      @db._end
      nil
    end
    
    def fetch_columns 
      self.get_columns
      @column_names
    end
    
    def get_name 
      @name.to_s
    end
    
  haxe_me ["coopy", "SqlTable"]
  end

end
